function [fopt, ncopt] = bbNcut(W_in)
% function [fopt, ncopt] = bbNcut(W_in)
% branch and bound for Ncut
%
% output:
%   fopt:     best partition f_n
%   ncopt:    Ncut value of f_n 
%
% input:
%   W_in:     m x m similarity matrix (contracted)

global W;
global volG;
global d;

% degrees and volume of the graph
d = sum(W_in);
volG = sum(d);


% heuristic: sort nodes by degree
[d, inds] = sort(d, 'descend');
W = W_in(inds,inds);
n = length(d);

singles = volG * (d - diag(W)')./(d.*(volG - d));

% set initial upper bound theta_u to the best partition of those separating 
% one node
[thu, ii] = min(singles);
fopt = zeros(n,1);
fopt(ii) = 1;

% start with f(1)=1, f(j)=0 for j>1
% (separating node 1)
f = zeros(n,1);
f(1) = 1;
iS = 2;
curr = singles(1);

% optimize and set fopt to the best solution
[f, ncopt] = optim(f, iS, thu, curr);

% if the returned solution is better than the single-node separation, then
% use it, otherwise not
if ncopt < thu
    fopt = f;
end

% re-sort indices
[trsh, indsi] = sort(inds);
fopt = fopt(indsi);

clear global W;

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function [fb, bcut] = optim(f, iS, minub, curr)
% find best possible assignment (by Ncut) of nodes {iS,...,n}
%
% f:     current assignment, fixed up to (exclusive) iS
% minub: theta_u, upper bound on best Ncut so far
% curr:  Ncut of current assignment f
%
% fb:   best partition found
% bcut: cut value of fb

global W d volG;

fx = f(1:(iS-1)); % A
fb = f; bcut = curr;

% base: iS==n
if iS == length(f)
    if min(fx)
        % f(i)=1 for all i in A, i.e. cluster 0 consists of one node only
        % that we hence cannot move
        return;
    else
        % return better of f(iS)=1 and f(iS)=0
        f(iS) = 1;
        bc = ncut(f);
        if bc < curr
            fb = f;
            bcut = bc;
        end
    end
    return;
end


% Pruning: would a further merge make sense at all?
% if so, test cases 0 and 1

% volume term: is better balancing possible at all?
wf = sum(W*f); % vol(cluster 1)
bbal = (wf < (0.5*volG));

% cut term: each node in undet. set has more nodes to 0-set than to 1-set
fi = ones(iS-1,1)-fx;
edgesums = (-fi + fx)'*W(1:(iS-1),iS:end);


% lower bound...
% cut term
if max(fi) % A_0 not empty
    lb = min( W(iS:end, 1:(iS-1))*fi );
    lb = fx'*W(1:(iS-1),1:(iS-1))*fi + lb;
else
    % we must assign at least one point to cluster 0
    % lb = min_(v in B) w(v,A_1)
    lb = min( W(iS:end, 1:(iS-1))*f(1:(iS-1)) );
end

% volume term
if bbal
    lb = lb * 4/volG;
else
    wf = wf + min(d(iS:end));
    lb = lb*(1/wf + 1/(volG-wf) );
end


if ((lb <= minub) && (max(edgesums)>0 || bbal))

    % continue, maybe improvement is possible
    % heuristic: alternately branch to 0 and 1 first
    if ~mod(iS,2)
        
        [fc, bc] = optim(f, iS+1, minub, curr);

        if bc < bcut
            fb = fc;
            bcut = bc;
            minub = min([bc minub]);
        end
        
        f(iS) = 1;
        currn = ncut(f);
        minub = min([currn minub]);
        [fc, bc] = optim(f, iS+1, minub, currn);

        if bc < bcut
            fb = fc;
            bcut = bc;
            minub = min([bc minub]);
        end
    else

        f(iS) = 1;
        currn = ncut(f);
        minub = min([currn minub]);
        [fc, bc] = optim(f, iS+1, minub, currn);

        if bc < bcut
            fb = fc;
            bcut = bc;
            minub = min([bc minub]);
        end
        
        f(iS) = 0;
        [fc, bc] = optim(f, iS+1, minub, curr);

        if bc < bcut
            fb = fc;
            bcut = bc;
            minub = min([bc minub]);
        end

    end

end

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%


function r = ncut(f)
% f has 0 and 1 entries
global W

n = length(f);
fi = ones(n,1)-f;

vA = sum(W)*f;
vB = sum(W)*fi;

if vA==0 || vB == 0
    r = Inf;
    return;
end

r = f'*W*fi * (1/vA + 1/vB);
